/*
 *    Amiga Linux/m68k Loader
 *
 *     Copyright 1995 by Geert Uytterhoeven
 *                       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *    --------------------------------------------------------------------------
 *
 *    The original idea for this program comes from the Linux distribution:
 *
 *       LILO - Generic Boot Loader for Linux ("LInux LOader")
 *              by Werner Almesberger
 *
 *    --------------------------------------------------------------------------
 *
 *    This file is subject to the terms and conditions of the GNU General Public
 *    License.  See the file README.legal in the main directory of the
 *    Linux/m68k distribution for more details.
 */


%{

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#include "lilo.h"


extern int yylex(void);
extern void yyerror(char *s);
extern int line;


struct ConfHeader ConfHeader = {
   NULL, NULL, 0, NULL, FALSE, 20, FALSE, 0, NULL, FALSE,
   0, 0, 0, 0, 0, 0, 0, 0, 0
};

static struct ConfBoot ConfBoot;


static void InitBoot()
{
   int i;

   ConfBoot.Next = NULL;
   ConfBoot.Label = NULL;
   ConfBoot.Type = BOS_LINUX;
   ConfBoot.Image = NULL;
   ConfBoot.Args = NULL;
   ConfBoot.Password = NULL;
   ConfBoot.ChipSize = 0;
   for (i = 0; i < arraysize(ConfBoot.FastAddress); i++) {
      ConfBoot.FastAddress[i] = 0;
      ConfBoot.FastSize[i] = 0;
   }
   ConfBoot.NumFastRAM = 0;
   ConfBoot.Model = AMI_UNKNOWN;
   ConfBoot.fType = 0;
   ConfBoot.fImage = 0;
   ConfBoot.fArgs = 0;
   ConfBoot.fPassword = 0;
   ConfBoot.fChipRAM = 0;
   ConfBoot.fModel = 0;
}


static void Redefinition(char *name)
{
   printf("%s:%d: Redefinition of `%s'\n", ConfigFile, line, name);
}

%}

%start file

%token HEADER ENDHEADER BOOTREC ENDREC BOOTDEV ALTDEV DEFAULT AUTO TIMEOUT AUX
       BAUD PASSWORD DEBUG TYPE MODEL IMAGE ARGS CHIPRAM FASTRAM STRING NUMBER
       BOOLEAN BOOTOS AMIMODEL

%%

file     :  header bootrecs
         ;

header   :  HEADER hdropts ENDHEADER
            {
               if (!ConfHeader.fBootDevice)
                  yyerror("No boot device specified");
               InitBoot();
            }
         ;

hdropts  :  /* empty */
         |  hdropts bootdev
         |  hdropts altdev
         |  hdropts default
         |  hdropts auto
         |  hdropts timeout
         |  hdropts aux
         |  hdropts baud
         |  hdropts mrpasswd
         |  hdropts debug
         ;

bootdev  :  BOOTDEV STRING
            {
               if (ConfHeader.fBootDevice)
                  Redefinition((char *)$1);
               ConfHeader.BootDevice = (char *)$2;
               ConfHeader.fBootDevice = 1;
            }
         ;

altdev   :  ALTDEV STRING NUMBER
            {
               if (ConfHeader.fAltDevice)
                  Redefinition((char *)$1);
               ConfHeader.AltDeviceName = (char *)$2;
               ConfHeader.AltDeviceUnit = $3;
               if (strlen(ConfHeader.AltDeviceName) >= MAXALTDEVSIZE)
                  yyerror("Alternate boot device name is too long");

               ConfHeader.fAltDevice = 1;
            }
         ;

default  :  DEFAULT STRING
            {
               if (ConfHeader.fDefault)
                  Redefinition((char *)$1);
               ConfHeader.Default = (char *)$2;
               ConfHeader.fDefault = 1;
            }
         ;

auto     :  AUTO BOOLEAN
            {
               if (ConfHeader.fAuto)
                  Redefinition((char *)$1);
               ConfHeader.Auto = $2;
               ConfHeader.fAuto = 1;
            }
         ;

timeout  :  TIMEOUT NUMBER
            {
               if (ConfHeader.fTimeOut)
                  Redefinition((char *)$1);
               ConfHeader.TimeOut = $2;
               ConfHeader.fTimeOut = 1;
            }
         ;

aux      :  AUX BOOLEAN
            {
               if (ConfHeader.fAux)
                  Redefinition((char *)$1);
               ConfHeader.Aux = $2;
               ConfHeader.fAux = 1;
            }
         ;

baud     :  BAUD NUMBER
            {
               if (ConfHeader.fBaud)
                  Redefinition((char *)$1);
               ConfHeader.Baud = $2;
               ConfHeader.fBaud = 1;
            }
         ;

mrpasswd :  PASSWORD STRING
            {
               if (ConfHeader.fPassword)
                  Redefinition((char *)$1);
               ConfHeader.Password = (char *)$2;
               if (strlen(ConfHeader.Password) >= MAXPWDSIZE)
                  yyerror("Password is too long");
               ConfHeader.fPassword = 1;
            }
         ;

debug    :  DEBUG BOOLEAN
            {
               if (ConfHeader.fDebug)
                  Redefinition((char *)$1);
               ConfHeader.Debug = $2;
               ConfHeader.fDebug = 1;
            }
         ;

bootrecs :  bootrec
         |  bootrecs bootrec
         ;

bootrec  :  BOOTREC STRING bootopts ENDREC
            {
               ConfBoot.Label = (char *)$2;
               if (strlen(ConfBoot.Label) >= MAXLABELSIZE)
                  yyerror("Boot record label is too long");
               if (ConfBoot.Type == BOS_AMIGA) {
                  if (ConfBoot.fImage)
                     yyerror("You can't specify an image for AmigaOS");
                  if (ConfBoot.fArgs)
                     yyerror("You can't specify arguments for AmigaOS");
               } else if (!ConfBoot.fImage)
                  yyerror("No boot image specified");
               if (!AddBootEntry(&ConfBoot))
                  yyerror("Duplicate boot image label");
               InitBoot();
            }
         ;

bootopts :  /* empty */
         |  bootopts type
         |  bootopts image
         |  bootopts args
         |  bootopts passwd
         |  bootopts chipram
         |  bootopts fastram
         |  bootopts model
         |  bootopts modelnum
         ;

type     :  TYPE BOOTOS
            {
               if (ConfBoot.fType)
                  Redefinition((char *)$1);
               ConfBoot.Type = $2;
               ConfBoot.fType = 1;
            }
         ;

image    :  IMAGE STRING
            {
               if (ConfBoot.fImage)
                  Redefinition((char *)$1);
               ConfBoot.Image = (char *)$2;
               ConfBoot.fImage = 1;
            }
         ;

args     :  ARGS STRING
            {
               if (ConfBoot.fArgs)
                  Redefinition((char *)$1);
               ConfBoot.Args = (char *)$2;
               if (strlen(ConfBoot.Args) >= CL_SIZE)
                  yyerror("Kernel argument is too long");
               ConfBoot.fArgs = 1;
            }
         ;

passwd   :  PASSWORD STRING
            {
               if (ConfBoot.fPassword)
                  Redefinition((char *)$1);
               ConfBoot.Password = (char *)$2;
               if (strlen(ConfBoot.Password) >= MAXPWDSIZE)
                  yyerror("Password is too long");
               ConfBoot.fPassword = 1;
            }
         ;

chipram  :  CHIPRAM NUMBER
            {
               if (ConfBoot.fChipRAM)
                  Redefinition((char *)$1);
               ConfBoot.ChipSize = $2;
               ConfBoot.fChipRAM = 1;
            }
         ;

fastram  :  FASTRAM NUMBER NUMBER
            {
               if (ConfBoot.NumFastRAM >= arraysize(ConfBoot.FastAddress))
                  yyerror("Too many FastRAM definitions");
               ConfBoot.FastAddress[ConfBoot.NumFastRAM] = $2;
               ConfBoot.FastSize[ConfBoot.NumFastRAM] = $3;
               ConfBoot.NumFastRAM++;
            }

model    :  MODEL AMIMODEL
            {
               if (ConfBoot.fModel)
                  Redefinition((char *)$1);
               ConfBoot.Model = $2;
               ConfBoot.fModel = 1;
            }
         ;

modelnum :  MODEL NUMBER
            {
               if (ConfBoot.fModel)
                  Redefinition((char *)$1);
               ConfBoot.Model = $2;
               ConfBoot.fModel = 1;
            }
         ;

%%
