/*
 *    Amiga Linux/m68k Loader
 *
 *     Copyright 1995 by Geert Uytterhoeven
 *                       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *    --------------------------------------------------------------------------
 *
 *    The original idea for this program comes from the Linux distribution:
 *
 *       LILO - Generic Boot Loader for Linux ("LInux LOader")
 *              by Werner Almesberger
 *
 *    --------------------------------------------------------------------------
 *
 *    This file is subject to the terms and conditions of the GNU General Public
 *    License.  See the file README.legal in the main directory of the
 *    Linux/m68k distribution for more details.
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/hdreg.h>
#include <asm/amigatypes.h>
#include <asm/bootinfo.h>

#include "lilo.h"


   /*
    *    Command Line Options
    */

char *ProgramName = NULL;
int Verbose = 0;
int Install = 1;
int Force = 0;
int DoChecksum = 0;
char *DosType = NULL;
char *SaveBootBlock = NULL;
char *RestoreBootBlock = NULL;
char *Device = NULL;


   /*
    *    Boot Code Definitions
    */

extern char _BootCodeStart[];
extern char _BootCodeEnd[];
extern char _LoaderVectorStart[];
extern char _LoaderVectorEnd[];
extern char _AltDeviceName[];
extern char _AltDeviceUnit[];


   /*
    *    Valid DosTypes
    */

struct DosTypeEntry {
   u_long ID;
   char *ASCII;
   char *Description;
};

struct DosTypeEntry DosTypeEntries[] = {
   { ID_BOOT_DISK, "BOOU", "Generic boot disk" },
   { ID_OFS_DISK, "DOS0", "OFS disk" },
   { ID_FFS_DISK, "DOS1", "FFS disk" },
   { ID_OFS_INTL_DISK, "DOS2", "OFS INTL disk" },
   { ID_FFS_INTL_DISK, "DOS3", "FFS INTL disk" },
   { ID_OFS_DC_DISK, "DOS4", "OFS DC disk" },
   { ID_FFS_DC_DISK, "DOS5", "FFS DC disk" },
   { ID_MUFS_DISK, "muFS", "MultiUser FFS INTL disk" },
   { ID_MU_OFS_DISK, "muF0", "MultiUser OFS disk" },
   { ID_MU_FFS_DISK, "muF1", "MultiUser FFS disk" },
   { ID_MU_OFS_INTL_DISK, "muF2", "MultiUser OFS INTL disk" },
   { ID_MU_FFS_INTL_DISK, "muF3", "MultiUser FFS INTL disk" },
   { ID_MU_OFS_DC_DISK, "muF4", "MultiUser OFS DC disk" },
   { ID_MU_FFS_DC_DISK, "muF5", "MultiUser FFS DC disk" },
   { 0, NULL, NULL }
};


   /*
    *    Lilo Data
    */

struct BootBlock BootBlock;

struct MapHeader MapHeader;
struct MapEntry *MapEntries = NULL;
u_long *MapVector;
int MapNumBlocks;

char *LoaderData;
int LoaderSize;
u_long *LoaderVector;
int LoaderNumBlocks;
u_long DosTypeID;

dev_t BootDevice;
struct stat info;
DIR *dir;
struct dirent *dirent;


   /*
    *    Alternate Boot Device Information
    */

char *AltDeviceName;
u_long AltDeviceUnit;


   /*
    *    Configuration File Parsing
    */

char ConfigFile[] = LILO_CONFIGFILE;

struct ConfBoot *BootEntries = NULL;


   /*
    *    Function Prototypes
    */

void Die(const char *fmt, ...) __attribute__ ((noreturn));
int AddBootEntry(struct ConfBoot *boot);
int main(int argc, char *argv[]);

static void ReadBootBlock(char *name);
static void WriteBootBlock(char *name);
static int ReadFile(char *name, void **data);
static void WriteFile(char *name, void *data, int size);
static void ReadConfigFile(void);
static void CreateMapData(void);
static void WriteMapData(void);
static u_long CalcChecksum(void);
static void Geometry(char *name, dev_t *device, u_long *start, u_long *blksize);
static u_long *CreateVector(char *name, int *numblocks);
static void PatchLoader(void);
static void CreateBootBlock(void);
static void SetDosType(void);
static void FixChecksum(void);
static void Usage(void) __attribute__ ((noreturn));


   /*
    *    Print an Error Message and Exit
    */

void Die(const char *fmt, ...)
{
   va_list ap;

   fflush(stdout);
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   va_end(ap);
   exit(1);
}


   /*
    *    Read a Boot Block from a Device or File
    */

static void ReadBootBlock(char *name)
{
   int fh;

   if ((fh = open(name, O_RDONLY)) == -1)
      Die("open %s: %s\n", name, strerror(errno));
   if (read(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
      Die("read %s: %s\n", name, strerror(errno));
   close(fh);
}


   /*
    *    Write a Boot Block to a Device
    */

static void WriteBootBlock(char *name)
{
   int fh;

   if ((fh = open(name, O_RDWR)) == -1)
      Die("open %s: %s\n", name, strerror(errno));
   if (write(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
      Die("write %s: %s\n", name, strerror(errno));
   close(fh);
}


   /*
    *    Read a File
    */

static int ReadFile(char *name, void **data)
{
   int fh, size = 0;

   if ((fh = open(name, O_RDONLY)) == -1)
      Die("open %s: %s\n", name, strerror(errno));
   if ((size = lseek(fh, 0, L_XTND)) == -1)
      Die("lseek %s: %s\n", name, strerror(errno));
   if (!(*data = malloc(size)))
      Die("No memory\n");
   if (lseek(fh, 0, L_SET) == -1)
      Die("lseek %s: %s\n", name, strerror(errno));
   if (read(fh, *data, size) != size)
      Die("read %s: %s\n", name, strerror(errno));
   close(fh);
   return(size);
}


   /*
    *    Write a File
    */

static void WriteFile(char *name, void *data, int size)
{
   int fh;

   if ((fh = open(name, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
      Die("open %s: %s\n", name, strerror(errno));
   if (write(fh, data, size) != size)
      Die("write %s: %s\n", name, strerror(errno));
   close(fh);
}


   /*
    *    Read the Configuration File
    */

static void ReadConfigFile(void)
{
   if (!(yyin = fopen(ConfigFile, "r")))
      Die("fopen %s: %s\n", ConfigFile, strerror(errno));
   yyparse();
   fclose(yyin);

   Device = ConfHeader.BootDevice;
   AltDeviceName = ConfHeader.AltDeviceName;
   AltDeviceUnit = ConfHeader.AltDeviceUnit;
   if (!ConfHeader.Default)
      ConfHeader.Default = BootEntries->Label;
}


   /*
    *    Add a New Boot Entry
    */

int AddBootEntry(struct ConfBoot *boot)
{
   struct ConfBoot **p, *new;

   for (p = &BootEntries; *p; p = &(*p)->Next)
      if (!strcmp((*p)->Label, boot->Label))
         return(0);
   if (!(new = malloc(sizeof(struct ConfBoot))))
      Die("No memory\n");
   *new = *boot;
   *p = new;
   return(1);
}


   /*
    *    Create the Map Data
    */

static void CreateMapData(void)
{
   struct ConfBoot *boot;
   struct MapEntry *entry;
   struct MapEntry **p;
   int size;
   int i = 0, j;
   u_long *vector;
   int numblocks;
   long defaultrec = -1;

   MapHeader.LiloID = LILO_ID;
   MapHeader.Flags = 0;
   if (ConfHeader.Auto)
      MapHeader.Flags |= LILOF_AUTO;
   if (ConfHeader.Debug)
      MapHeader.Flags |= LILOF_DEBUG;
   if (ConfHeader.Aux)
      MapHeader.Flags |= LILOF_AUX;
   MapHeader.TimeOut = ConfHeader.TimeOut;
   if (ConfHeader.Password)
      strcpy(MapHeader.MasterPwd, ConfHeader.Password);
   else
      strcpy(MapHeader.MasterPwd, "");
   MapHeader.Baud = ConfHeader.Baud;

   boot = BootEntries;
   p = &MapEntries;

   do {
      if (!strcmp(ConfHeader.Default, boot->Label))
         defaultrec = i;
      size = 0;
      if (boot->Type == BOS_AMIGA)
         vector = NULL;
      else if ((vector = CreateVector(boot->Image, &numblocks)))
         size = (numblocks+1)*sizeof(u_long);
      else if (i == defaultrec)
         Die("Image `%s' for default boot record `%s' does not exist\n",
             boot->Image, boot->Label);
      else
         size = 0;
      entry = malloc(sizeof(struct MapEntry)+size);
      entry->Size = sizeof(struct MapEntry)+size;
      strcpy(entry->Label, boot->Label);
      if (!boot->Password)
         strcpy(entry->Password, "");
      else if (i == defaultrec)
         Die("Default boot record `%s' must not have a password\n",
             boot->Label);
      else
         strcpy(entry->Password, boot->Password);
      if (vector)
         entry->Type = boot->Type;
      else
         entry->Type = -boot->Type;
      if (boot->Args)
         strcpy(entry->CommandLine, boot->Args);
      else
         strcpy(entry->CommandLine, "");
      if (vector) {
         memcpy(entry->KernelVector, vector, size);
         free(vector);
      }
      entry->ChipSize = boot->ChipSize;
      for (j = 0; j < arraysize(boot->FastAddress); j++) {
         entry->FastAddress[j] = boot->FastAddress[j];
         entry->FastSize[j] = boot->FastSize[j];
      }
      entry->NumFastRAM = boot->NumFastRAM;
      entry->Model = boot->Model;
      *p = entry;
      p = &entry->un.Next;
      i++;
   } while ((boot = boot->Next));

   if (defaultrec == -1)
      Die("Default boot record `%s' does not exist\n", ConfHeader.Default);
   MapHeader.Default = defaultrec;

   entry->un.Next = NULL;
}


   /*
    *    Write the Map Data
    */

static void WriteMapData(void)
{
   int fh, i = 0;
   struct MapEntry *entry, *next;

   if ((fh = open(LILO_MAPFILE, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
      Die("open " LILO_MAPFILE ": %s\n", strerror(errno));
   if (write(fh, &MapHeader, sizeof(struct MapHeader)) !=
       sizeof(struct MapHeader))
      Die("write " LILO_MAPFILE ": %s\n", strerror(errno));
   entry = MapEntries;
   puts("Boot records:");
   do {
      next = entry->un.Next;
      entry->un.NotLast = next ? 1 : 0;
      if (i == MapHeader.Default)
         printf("[1m");
      printf("   %-32.32s (", entry->Label);
      if (entry->Type > 0)
         printf("%lu bytes)\n", entry->KernelVector[0]);
      else if (entry->Type == BOS_AMIGA)
         puts("ROM image)");
      else
         puts("not available)");
      if (i == MapHeader.Default)
         printf("[0m");
      if (write(fh, entry, entry->Size) != entry->Size)
         Die("write " LILO_MAPFILE ": %s\n", strerror(errno));
      i++;
   } while ((entry = next));
   close(fh);
}


   /*
    *    Calculate the Boot Block Checksum
    */

static u_long CalcChecksum(void)
{
   u_long _res;
   struct BootBlock *bb = &BootBlock;

   __asm __volatile ("1: addl %0@+,%1;"
                     "   addxl %5,%1;"
                     "   dbf %4,1b;"
                     : "=a" (bb), "=d" (_res)
                     : "0" (bb), "1" (0), "d" (255), "d" (0));
   return(~_res);
}


   /*
    *    Get the Geometry for the Partition a file belongs to
    */

static void Geometry(char *name, dev_t *device, u_long *start, u_long *blksize)
{
   struct hd_geometry geo;
   char devname[PATH_MAX+1];
   int fh, found = 0;

   if (stat(name, &info) == -1)
      Die("stat %s: %s\n", name, strerror(errno));
   if (!S_ISREG(info.st_mode))
      Die("%s is not a regular file\n", name);
   *device = info.st_dev;
   *blksize = info.st_blksize;

   if (!(dir = opendir("/dev")))
      Die("opendir /dev: %s\n", strerror(errno));
   while ((dirent = readdir(dir))) {
      sprintf(devname, "/dev/%s", dirent->d_name);
      if ((stat(devname, &info) != -1) && S_ISBLK(info.st_mode) &&
          (info.st_rdev == *device)) {
         if ((fh = open(devname, O_RDONLY)) == -1)
            Die("open %s: %s\n", devname, strerror(errno));
         if (ioctl(fh, HDIO_GETGEO, &geo) == -1)
            Die("ioctl %s: %s\n", devname, strerror(errno));
         close(fh);
         found = 1;
         break;
      }
   }
   closedir(dir);
   if (found)
      *start = 512*geo.start;
   else
      Die("Can't find special device entry for file `%s'\n");
}


   /*
    *    Create a Vector for a File
    */

static u_long *CreateVector(char *name, int *numblocks)
{
   int fh, size, i;
   u_long *vector = NULL;
   u_long offset, start, blksize;
   dev_t device;
   int relblksize;
   static u_int first = 1;
   static dev_t altdev;

   if (!name || ((fh = open(name, O_RDONLY)) == -1))
      return(NULL);
   if ((size = lseek(fh, 0, L_XTND)) == -1)
      Die("lseek %s: %s\n", name, strerror(errno));
   *numblocks = (size+LILO_BLOCKSIZE-1)/LILO_BLOCKSIZE;
   if (!(vector = malloc((*numblocks+1)*sizeof(u_long))))
      Die("No memory\n");
   vector[0] = size;

   Geometry(name, &device, &start, &blksize);

   if (!AltDeviceName) {
      if (MAJOR(device) != MAJOR(BootDevice))
         Die("File `%s' must reside on physical device `%s'\n", name, Device);
      switch (MAJOR(BootDevice)) {
         case FLOPPY_MAJOR:
            if (MINOR(device) != MINOR(BootDevice))
               Die("File `%s' must reside on physical floppy device `%s'\n",
                   name, Device);
            break;
         case HD_MAJOR:
            if ((MINOR(device) & 0xc0) != (MINOR(BootDevice) & 0xc0))
               Die("File `%s' must reside on physical IDE device `%s'\n", name,
                   Device);
            break;
         case SCSI_DISK_MAJOR:
            if ((MINOR(device) & 0xf0) != (MINOR(BootDevice) & 0xf0))
               Die("File `%s' must reside on physical SCSI device `%s'\n", name,
                   Device);
            break;
         default:
            Die("Major device %d not supported (yet)\n", MAJOR(device));
            break;
      }
   } else if (first) {
      altdev = device;
      printf("Assuming `%s' resides on `%s:%ld'\n", name, AltDeviceName,
             AltDeviceUnit);
      first = 0;
   } else {
      if (MAJOR(device) != MAJOR(altdev))
         Die("File `%s' must reside on `%s:%ld'\n", name, AltDeviceName,
             AltDeviceUnit);
      switch (MAJOR(altdev)) {
         case FLOPPY_MAJOR:
            if (MINOR(device) != MINOR(altdev))
               Die("File `%s' must reside on physical floppy device `%s:%ld'\n",
                   name, AltDeviceName, AltDeviceUnit);
            break;
         case HD_MAJOR:
            if ((MINOR(device) & 0xc0) != (MINOR(altdev) & 0xc0))
               Die("File `%s' must reside on physical IDE device `%s:%ld'\n",
                   name, AltDeviceName, AltDeviceUnit);
            break;
         case SCSI_DISK_MAJOR:
            if ((MINOR(device) & 0xf0) != (MINOR(altdev) & 0xf0))
               Die("File `%s' must reside on physical SCSI device `%s:%ld'\n",
                   name, AltDeviceName, AltDeviceUnit);
            break;
         default:
            Die("Major device %d not supported (yet)\n", MAJOR(device));
            break;
      }
   }

   if (blksize % LILO_BLOCKSIZE)
      Die("File `%s' resides on a device with an unsupported block size (%d bytes)\n",
          name, blksize);

   relblksize = blksize/LILO_BLOCKSIZE;

   for (i = 0; i < *numblocks; i++) {
      if (!(i % relblksize)) {
         offset = i/relblksize;
         if (ioctl(fh, FIBMAP, &offset) == -1)
            Die("ioctl FIBMAP %s: %s\n", name, strerror(errno));
         if (offset)
            offset = start+offset*blksize;
      } else if ((offset = vector[i]))
         offset += LILO_BLOCKSIZE;
      vector[i+1] = offset;
   }
   close(fh);
   return(vector);
}


   /*
    *    Patch the Loader with the Map Vector
    */

static void PatchLoader(void)
{
   int i;
   u_long pattern[] = {
      LILO_ID, LILO_MAPVECTORID
   };
   u_long *data = NULL;
   u_long maxsize;

   for (i = 0; i < LoaderSize-sizeof(pattern)+1; i++)
      if (!strncmp(&LoaderData[i], (char *)pattern, sizeof(pattern))) {
         data = (u_long *)&LoaderData[i];
         break;
      }
   if (!data)
      Die("Couldn't find magic key in loader template\n");
   maxsize = data[2];
   if (MapNumBlocks > maxsize-1)
      Die("Map vector is too large for loader (%d entries too much)\n",
          MapNumBlocks-maxsize-1);
   if (Verbose)
      printf("%d entries for map vector blocks, %lu entries free\n",
             MapNumBlocks, maxsize-1-MapNumBlocks);
   memcpy(&data[3], MapVector, (MapNumBlocks+1)*sizeof(u_long));
}


   /*
    *    Create the Boot Block
    */

static void CreateBootBlock(void)
{
   u_long maxsize, size;

   size = _BootCodeEnd-_BootCodeStart;
   if (size > sizeof(BootBlock.Data))
      Die("%d bytes too much boot code\n", size-sizeof(BootBlock.Data));
   if (Verbose)
      printf("%lu bytes boot code, %lu bytes free\n", size,
             sizeof(BootBlock.Data)-size);
   BootBlock.ID = DosTypeID;
   if (BootBlock.ID == ID_BOOT_DISK)
      BootBlock.DosBlock = 0;
   memcpy(BootBlock.Data, _BootCodeStart, size);

   maxsize = (_LoaderVectorEnd-_LoaderVectorStart)/sizeof(u_long);
   if (LoaderNumBlocks > maxsize-1)
      Die("Loader vector is too large for boot block (%d entries too much)\n",
          LoaderNumBlocks-maxsize-1);
   if (Verbose)
      printf("%d entries for loader vector blocks, %lu entries free\n",
             LoaderNumBlocks, maxsize-1-LoaderNumBlocks);
   memcpy((void *)((u_long)BootBlock.Data+(u_long)_LoaderVectorStart-
                   (u_long)_BootCodeStart), LoaderVector,
          (LoaderNumBlocks+1)*sizeof(u_long));
   if (AltDeviceName) {
      strcpy((void *)((u_long)BootBlock.Data+(u_long)_AltDeviceName-
                      (u_long)_BootCodeStart), AltDeviceName);
      memcpy((void *)((u_long)BootBlock.Data+(u_long)_AltDeviceUnit-
                      (u_long)_BootCodeStart), &AltDeviceUnit,
                      sizeof(AltDeviceUnit));
   }
   BootBlock.LiloID = LILO_ID;
}


   /*
    *    Set the DosType of the Boot Block
    */

static void SetDosType(void)
{
   int i;

   for (i = 0; DosTypeEntries[i].ID; i++)
      if (!strcmp(DosType, DosTypeEntries[i].ASCII))
         break;
   if (DosTypeEntries[i].ID)
      DosTypeID = DosTypeEntries[i].ID;
   else {
      printf("Unknown DosType `%s'\n", DosType);
      printf("Valid DosTypes are:\n");
      for (i = 0; DosTypeEntries[i].ID; i++)
         printf(" %s", DosTypeEntries[i].ASCII);
      Die("\n");
   }
   BootBlock.ID = DosTypeID;
}


   /*
    *    Fix the Checksum of the Boot Block
    */

static void FixChecksum(void)
{
   BootBlock.Checksum = 0;
   BootBlock.Checksum = CalcChecksum();
}


   /*
    *    Print the Usage Template and Exit
    */

static void Usage(void)
{
   Die("Usage: %s [options]\n\n"
       "Valid options are:\n"
       "   -h, --help                : Display this usage information\n"
       "   -v, --verbose             : Enable verbose mode\n"
       "   -f, --force               : Force installation over an unknown boot block\n"
       "   -t, --dostype <dostype>   : Dostype for the boot block\n"
       "   -d, --device <device>     : Block special device for restore/save/checksum\n"
       "   -r, --restore-from <file> : File to restore the boot block from\n"
       "   -s, --save-to <file>      : File to save the boot block to\n"
       "   -c, --checksum            : Fix the boot block checksum\n\n",
       ProgramName);
}


   /*
    *    Main Routine
    */

int main(int argc, char *argv[])
{
   int i;

   ProgramName = argv[0];
   argc--;
   argv++;
   puts(VERSION "\n");

   while (argc) {
      if (!strcmp(argv[0], "-h") || !strcmp(argv[0], "--help"))
         Usage();
      if (!strcmp(argv[0], "-v") || !strcmp(argv[0], "--verbose"))
         Verbose = 1;
      else if (!strcmp(argv[0], "-f") || !strcmp(argv[0], "--force"))
         Force = 1;
      else if (!strcmp(argv[0], "-d") || !strcmp(argv[0], "--device"))
         if (--argc && !Device) {
            Device = argv[1];
            argv++;
         } else
            Usage();
      else if (!strcmp(argv[0], "-r") || !strcmp(argv[0], "--restore-from"))
         if (--argc && !RestoreBootBlock) {
            RestoreBootBlock = argv[1];
            Install = 0;
            argv++;
         } else
            Usage();
      else if (!strcmp(argv[0], "-s") || !strcmp(argv[0], "--save-to"))
         if (--argc && !SaveBootBlock) {
            SaveBootBlock = argv[1];
            Install = 0;
            argv++;
         } else
            Usage();
      else if (!strcmp(argv[0], "-t") || !strcmp(argv[0], "--dostype"))
         if (--argc && !DosType) {
            DosType = argv[1];
            argv++;
         } else
            Usage();
      else if (!strcmp(argv[0], "-c") || !strcmp(argv[0], "--device")) {
         Install = 0;
         DoChecksum = 1;
      } else
         Usage();
      argc--;
      argv++;
   }

   if (Device && Install && !DoChecksum)
      Die("You can specify a device only with -s, -r or -c\n");
   if (!Device && !Install)
      Die("You must specify a device if you use -s, -r or -c\n");

   if (Install)
      ReadConfigFile();

   if (stat(Device, &info) == -1)
      Die("stat %s: %s\n", Device, strerror(errno));
   if (!S_ISBLK(info.st_mode))
      Die("%s is not a block special device\n", Device);
   BootDevice = info.st_rdev;

   ReadBootBlock(Device);
   printf("%s: ", Device);
   DosTypeID = BootBlock.ID;
   for (i = 0; DosTypeEntries[i].ID; i++)
      if (BootBlock.ID == DosTypeEntries[i].ID)
         break;
   if (DosTypeEntries[i].ID)
      printf("%s", DosTypeEntries[i].Description);
   else {
      printf("Type 0x%08lx", BootBlock.ID);
      if (!Force) {
         printf("\n");
         Die("Use option -f to force continue\n");
      }
      DosTypeID = ID_BOOT_DISK;
   }
   printf("%s, Boot block checksum is %s\n",
          BootBlock.LiloID == LILO_ID ? " (created by Lilo)" : "",
          CalcChecksum() ? "Incorrect" : "OK");

   if (SaveBootBlock) {
      WriteFile(SaveBootBlock, &BootBlock, sizeof(BootBlock));
      printf("Boot block saved to file `%s'\n", SaveBootBlock);
   }
   if (RestoreBootBlock) {
      ReadBootBlock(RestoreBootBlock);
      printf("Boot block restored from file `%s'\n", RestoreBootBlock);
   }
   if (DosType)
      SetDosType();
   if (Install) {
      CreateMapData();
      WriteMapData();
      LoaderSize = ReadFile(LILO_LOADERTEMPLATE, (void **)&LoaderData);
      if (!(MapVector = CreateVector(LILO_MAPFILE, &MapNumBlocks)))
         Die("open `" LILO_MAPFILE "': %s\n", strerror(errno));
      if (Verbose)
         printf("Map vector has %lu bytes\n", MapVector[0]);
      PatchLoader();
      WriteFile(LILO_LOADERFILE, LoaderData, LoaderSize);
      if (!(LoaderVector = CreateVector(LILO_LOADERFILE, &LoaderNumBlocks)))
         Die("open `" LILO_LOADERFILE "': %s\n", strerror(errno));
      if (Verbose)
         printf("Loader vector has %lu bytes\n", LoaderVector[0]);
      CreateBootBlock();
   }
   if (Install || DoChecksum)
      FixChecksum();
   if (RestoreBootBlock || Install || DoChecksum) {
      WriteBootBlock(Device);
      printf("Boot block written to device `%s'\n", Device);
   }
   exit(0);
}
