/*
 *    Amiga Linux/m68k Loader
 *
 *     Copyright 1995 by Geert Uytterhoeven
 *                       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *    --------------------------------------------------------------------------
 *
 *    This program is based on files of the Linux/m68k distribution:
 *
 *     o bootstrap.c -- This program loads the Linux/m68k kernel into an Amiga
 *                      and launches it.
 *
 *       Copyright 1993, 1994 by Hamish Macdonald, Greg Harp
 *
 *       Modified 11-May-94 by Geert Uytterhoeven: A3640 MapROM check
 *       Modified 31-May-94 by Geert Uytterhoeven: Memory thrash problem solved
 *       Modified 11-July-95 by Andreas Schwab: Support for ELF kernel
 *
 *     o bootstrap.h -- This file is a part of the Amiga bootloader.
 *
 *       Copyright 1993, 1994 by Hamish Macdonald
 *
 *       Some minor additions by Michael Rausch 1-11-94
 *       Modified 11-May-94 by Geert Uytterhoeven: inline Supervisor() call
 *
 *
 *    The original idea for this program comes from the Linux distribution:
 *
 *       LILO - Generic Boot Loader for Linux ("LInux LOader")
 *              by Werner Almesberger
 *
 *    --------------------------------------------------------------------------
 *
 *    This file is subject to the terms and conditions of the GNU General Public
 *    License.  See the file README.legal in the main directory of the
 *    Linux/m68k distribution for more details.
 *
 *    --------------------------------------------------------------------------
 *
 *    WARNING: Because this code must be PC-relative, it's compiled with the
 *             -fPIC option and converted with my `fixpic' perl script.
 *             As a consequence we must not use pointers that are initialized at
 *             compile time but we must explicitly initialize them at run time.
 *             Look at the initialization of NewScreen.DefaultTitle and
 *             NewWindow.Title for an example.
 *
 *             NOTE: This applies to linuxboot.c too!
 */


#ifndef __GNUC__
#error GNU CC is required to compile this program
#endif /* __GNUC__ */


#define V36_COMPATIBILITY     /* Make it work with the A3000 Beta ROM */


#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <linux/a.out.h>
#include <linux/elf.h>
#include <asm/bootinfo.h>
#include <asm/amigatypes.h>
#include <asm/amigahw.h>

#include "linuxboot.h"
#include "loader.h"
#include "bootlib.h"
#include "lilo.h"


   /*
    *    Lilo Alert
    */

#define AN_LILO      (AT_DeadEnd|0x4c000000)    /* `L' */
#define AO_LiloMap   (0x004d4150)               /* `MAP': Corrupt Map File */


   /*
    *    Console Echo Modes
    */

#define ECHO_NONE       (0)
#define ECHO_NORMAL     (1)
#define ECHO_DOT        (2)


   /*
    *    Custom Hardware Bases
    *
    *    Make sure these are the physical and not the virtual address
    */

#undef custom
#define custom          ((*(volatile struct CUSTOM *)(CUSTOM_PHYSADDR)))
#undef ciaa
#define ciaa            ((*(volatile struct CIA *)(CIAA_PHYSADDR)))


   /*
    *    Boot Library Functions
    */

static struct LiloTable *LiloTable;


   /*
    *    Library Bases
    */

#ifdef V36_COMPATIBILITY
#define LIB_VERS        (36)
#else /* V36_COMPATIBILITY */
#define LIB_VERS        (37)
#endif /* V36_COMPATIBILITY */

struct ExecBase *SysBase;
struct ExpansionBase *ExpansionBase;
struct GfxBase *GfxBase;
static struct IntuitionBase *IntuitionBase;
static struct InputBase *InputBase;
#ifdef V36_COMPATIBILITY
static struct KeyboardBase *KeyboardBase;
#endif /* V36_COMPATIBILITY */


   /*
    *    Console Stuff
    */

static u_char Title[] = "Amiga Linux/m68k Loader";

static struct NewScreen NewScreen = {
   0, 0, 0, 0, 2, 0, 1, HIRES, CUSTOMSCREEN, NULL, NULL, NULL, NULL
};
static struct NewWindow NewWindow = {
   0, 0, 0, 0, 2, 3, 0, WFLG_SIMPLE_REFRESH|WFLG_ACTIVATE, NULL, NULL,
   NULL, NULL, NULL, 0, 0, 0, 0, CUSTOMSCREEN
};
static struct Screen *Screen;
static struct Window *Window;
static struct MsgPort *ConsolePort;
static struct IOStdReq *ConsoleRequest = NULL;
static struct MsgPort *TimerPort;
static struct timerequest *TimerRequest;
static struct MsgPort *InputPort;
static struct IOStdReq *InputRequest;
#ifdef V36_COMPATIBILITY
static struct MsgPort *KeyboardPort;
static struct IOStdReq *KeyboardRequest;
static u_char KeyboardMatrix[13];
#endif /* V36_COMPATIBILITY */
static u_long TimeOutSignalMask = 0;
static char FmtBuffer[256], InBuffer[CL_SIZE];
static u_long ConsoleOpened = FALSE;
static void *OldDebug;


   /*
    *    Map Data
    */

#define MAXVECTORSIZE   (1024)

struct FileVectorData {
   u_long ID[2];
   u_long MaxVectorSize;
   u_long Vector[MAXVECTORSIZE];
};

static struct FileVectorData MapVectorData = {
   { LILO_ID, LILO_MAPVECTORID }, MAXVECTORSIZE
};

static struct MapHeader MapHeader;
static struct MapEntry *MapEntries = NULL;
static u_short MasterMode = FALSE;
static u_short Auto = FALSE;
static u_short Aux = FALSE;
static u_short Debug = FALSE;


   /*
    *    Linux/m68k Bootstrap Stuff
    */

static struct MapEntry *BootEntry;


   /*
    *    Function Prototypes
    */

u_long Main(struct LiloTable *lilotable);
static void ReadMapData(void);
static void FreeMapData(void);
#ifdef V36_COMPATIBILITY
static u_short PeekQualifier36(void);
#endif /* V36_COMPATIBILITY */
static u_long ProbeConsole();
static void OpenConsole(void);
static void CloseConsole();
static long ReadLine(u_long mode);
static void SetTimeOut(u_long seconds);
static int TimedOut(void);
static void ClearTimeOut(void);
static struct MapEntry *GetDefaultBootOS(void);
static struct MapEntry *GetBootOS(void);
static u_long BootAmigaOS(void);


   /*
    *    Main Loader Routine
    *
    *    This routine should return FALSE in case of an error (which will cause
    *    a reboot), and TRUE if the user wants to boot AmigaOS.
    *    In all other cases it won't return :-)
    */

u_long Main(struct LiloTable *lilotable)
{
   u_long res = FALSE;

   /* Boot Library */
   LiloTable = lilotable;

   /* Open Libraries */

   SysBase = *(struct ExecBase **)4;

   if (!(ExpansionBase =
         (struct ExpansionBase *)OpenLibrary("expansion.library", LIB_VERS)))
      Alert(AN_LILO|AG_OpenLib|AO_ExpansionLib);
   if (!(GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", LIB_VERS)))
      Alert(AN_LILO|AG_OpenLib|AO_GraphicsLib);

   /* Open Timer Device */

   if (!(TimerPort = CreateMsgPort()))
      Alert(AN_LILO|AG_NoSignal);
   if (!(TimerRequest = CreateIORequest(TimerPort, sizeof(struct timerequest))))
      Alert(AN_LILO|AG_NoMemory);
   if (OpenDevice("timer.device", UNIT_VBLANK, (struct IORequest *)TimerRequest,
                  0))
      Alert(AN_LILO|AG_OpenDev|AO_TimerDev);

   /* Open Input Device */

   if (!(InputPort = CreateMsgPort()))
      Alert(AN_LILO|AG_NoSignal);
   if (!(InputRequest = CreateIORequest(InputPort, sizeof(struct IOStdReq))))
      Alert(AN_LILO|AG_NoMemory);
   if (OpenDevice("input.device", 0, (struct IORequest *)InputRequest, 0))
      Alert(AN_LILO|AG_OpenDev|AO_Unknown);
   InputBase = (struct InputBase *)InputRequest->io_Device;

   /* Configuration */

   ReadMapData();
   if (!Auto || ProbeConsole()) {
      OpenConsole();
      BootEntry = GetBootOS();
   } else
      BootEntry = GetDefaultBootOS();

   switch (BootEntry->Type) {
      case BOS_AMIGA:
         Printf("\nBooting AmigaOS (%s)...\n\n", BootEntry->Label);
         res = BootAmigaOS();
         break;

      case BOS_LINUX:
         Printf("\nBooting Linux/m68k (%s)...\n\n", BootEntry->Label);
         res = BootLinux(BootEntry->Label, NULL, BootEntry->CommandLine, Debug,
                         0);
         break;

      case BOS_NETBSD:
         Printf("\nBooting NetBSD-Amiga (%s)...\n\n", BootEntry->Label);
         Puts("Not yet implemented :-)\n");
         res = FALSE;
         break;
   }
   if (!res) {
      if (BootEntry->Type < 0)
         Puts("\nPANIC: Boot Operating System is not available\n");
      else
         Puts("\nPANIC: Couldn't boot desired Operating System\n");
      Sleep(3000000);
   }

   CloseConsole();

   FreeMapData();

   /* Close Input Device */

   CloseDevice((struct IORequest *)InputRequest);
   DeleteIORequest(InputRequest);
   DeleteMsgPort(InputPort);

   /* Close Timer Device */

   CloseDevice((struct IORequest *)TimerRequest);
   DeleteIORequest(TimerRequest);
   DeleteMsgPort(TimerPort);

   /* Close Libraries */

   CloseLibrary((struct Library*)GfxBase);
   CloseLibrary((struct Library *)ExpansionBase);

   return(res);
}


   /*
    *    Read the Map Data
    */

static void ReadMapData(void)
{
   u_long size;
   struct MapEntry *entry;
   int i = 0;

   Lilo_Open(MapVectorData.Vector);
   Lilo_Read(&MapHeader, sizeof(struct MapHeader));
   if (MapHeader.LiloID != LILO_ID)
      Alert(AN_LILO|AO_LiloMap);
   if (MapHeader.Flags & LILOF_AUTO)
      Auto = TRUE;
   if (MapHeader.Flags & LILOF_AUX) {
      Aux = TRUE;
      if (MapHeader.Baud)
         custom.serper = (5*SysBase->ex_EClockFrequency+MapHeader.Baud/2)/
                         MapHeader.Baud-1;
   }
   if (MapHeader.Flags & LILOF_DEBUG) {
      OpenConsole();
      Debug = TRUE;
      Printf("Auto:            %s\n", Auto ? "Enabled" : "Disabled");
      Printf("Debug:           %s\n", Debug ? "Enabled" : "Disabled");
      Printf("Aux:             %s\n", Aux ? "Enabled" : "Disabled");
      Printf("Default Image:   #%ld\n", MapHeader.Default);
      Printf("Time Out:        %ld\n", MapHeader.TimeOut);
      Printf("Master Password: `%s'\n", MapHeader.MasterPwd);
      if (MapHeader.Baud)
         Printf("Baud:            %ld\n", MapHeader.Baud);
      else
         Puts("Baud:            ROM default\n");
   }
   size = MapVectorData.Vector[0]-sizeof(struct MapHeader);
   if (Debug)
      Printf("Map file contains %ld bytes of data\n\n", size);
   if (!(MapEntries = (struct MapEntry *)AllocMem(size,
                                                  MEMF_PUBLIC|MEMF_CLEAR)))
      Alert(AN_LILO|AG_NoMemory);
   Lilo_Read(MapEntries, size);
   Lilo_Close();
   entry = MapEntries;
   do {
      if (entry->un.NotLast)
         entry->un.Next = (struct MapEntry *)((u_long)entry+entry->Size);
      else
         entry->un.Next = NULL;
      if (Debug) {
         if (i == MapHeader.Default)
            Puts("[1m");
         Printf("%-32.32s (", entry->Label);
         if (entry->Type > 0)
            Printf("%ld bytes, type %ld)\n", entry->KernelVector[0],
                   entry->Type);
         else if (entry->Type == BOS_AMIGA)
            Puts("ROM image)\n");
         else
            Puts("not available)\n");
         if (i == MapHeader.Default)
            Puts("[0m");
      }
      i++;
   } while ((entry = entry->un.Next));
}


   /*
    *    Free the Map Data
    */

static void FreeMapData(void)
{
   if (MapEntries)
      FreeMem(MapEntries, MapVectorData.Vector[0]-sizeof(struct MapHeader));
}


#ifdef V36_COMPATIBILITY
   /*
    *    V36 (A3000 Beta ROM) compatible replacement for PeekQualifier()
    */

static u_short PeekQualifier36(void)
{
   u_short keys;

   /* Open Keyboard Device */

   if (!(KeyboardPort = CreateMsgPort()))
      Alert(AN_LILO|AG_NoSignal);
   if (!(KeyboardRequest = CreateIORequest(KeyboardPort,
                                           sizeof(struct IOStdReq))))
      Alert(AN_LILO|AG_NoMemory);
   if (OpenDevice("keyboard.device", 0, (struct IORequest *)KeyboardRequest, 0))
      Alert(AN_LILO|AG_OpenDev|AO_TimerDev);
   KeyboardBase = (struct KeyboardBase *)KeyboardRequest->io_Device;

   KeyboardRequest->io_Command = KBD_READMATRIX;
   KeyboardRequest->io_Flags = IOF_QUICK;
   KeyboardRequest->io_Length = sizeof(KeyboardMatrix);
   KeyboardRequest->io_Data = KeyboardMatrix;
   DoIO((struct IORequest *)KeyboardRequest);

   keys = KeyboardMatrix[12];
   if (!(custom.potgor & 0x0100))
      keys |= IEQUALIFIER_MIDBUTTON;
   if (!(custom.potgor & 0x0400))
      keys |= IEQUALIFIER_RBUTTON;
   if (!(ciaa.pra & 0x40))
      keys |= IEQUALIFIER_LEFTBUTTON;

   /* Close Keyboard Device */

   CloseDevice((struct IORequest *)KeyboardRequest);
   DeleteIORequest(KeyboardRequest);
   DeleteMsgPort(KeyboardPort);

   return(keys);
}
#endif /* V36_COMPATIBILITY */


   /*
    *    Probe the Console
    */

static u_long ProbeConsole(void)
{
   u_long res;

#ifdef V36_COMPATIBILITY
   if (SysBase->Version >= 37)
#endif /* V36_COMPATIBILITY */
      res = PeekQualifier() ? TRUE : FALSE;
#ifdef V36_COMPATIBILITY
   else
      res = PeekQualifier36() ? TRUE : FALSE;
#endif /* V36_COMPATIBILITY */

   if (!res && Aux) {
      SetTimeOut(1);
      do
         if (RawMayGetChar() != -1) {
            res = TRUE;
            while (RawMayGetChar() != -1);
            break;
         }
      while (!TimedOut());
      ClearTimeOut();
   }
   return(res);
}


   /*
    *    Replacement for exec.library/Debug()
    *
    *    We don't want to meet a debugger when DEL is pressed on the terminal
    */

asm("
.text
NewDebug:      rts
");
extern void NewDebug(void);
 

   /*
    *    Open the Console
    */

static void OpenConsole(void)
{
   if (!ConsoleOpened) {
      if (!(IntuitionBase =
            (struct IntuitionBase *)OpenLibrary("intuition.library", LIB_VERS)))
         Alert(AN_LILO|AG_OpenLib|AO_Intuition);
      NewScreen.Width = GfxBase->NormalDisplayColumns;
      NewScreen.Height = GfxBase->NormalDisplayRows;
      NewScreen.DefaultTitle = Title;
      if (!(Screen = OpenScreen(&NewScreen)))
         Alert(AN_LILO|AG_OpenRes|AO_Intuition);
      NewWindow.Width = NewScreen.Width;
      NewWindow.Height = NewScreen.Height;
      NewWindow.Screen = Screen;
      NewWindow.Title = Title;
      if (!(Window = OpenWindow(&NewWindow)))
         Alert(AN_LILO|AG_OpenRes|AO_Intuition);
      if (!(ConsolePort = CreateMsgPort()))
         Alert(AN_LILO|AG_NoSignal);
      if (!(ConsoleRequest = CreateIORequest(ConsolePort,
                                             sizeof(struct IOStdReq))))
         Alert(AN_LILO|AG_NoMemory);
      ConsoleRequest->io_Data = Window;
      if (OpenDevice("console.device", CONU_SNIPMAP,
                     (struct IORequest *)ConsoleRequest, CONFLAG_DEFAULT))
         Alert(AN_LILO|AG_OpenDev|AO_ConsoleDev);
      if (Aux)
         OldDebug = SetFunction((struct Library *)SysBase, LVODebug, NewDebug);
      ConsoleOpened = TRUE;
      Puts(VERSION "\n\n");
   }
}


   /*
    *    Close the Console
    */

static void CloseConsole()
{
   if (ConsoleOpened) {
      if (Aux)
         SetFunction((struct Library *)SysBase, LVODebug, OldDebug);
      CloseDevice((struct IORequest *)ConsoleRequest);
      DeleteIORequest(ConsoleRequest);
      DeleteMsgPort(ConsolePort);
      CloseWindow(Window);
      CloseScreen(Screen);
      CloseLibrary((struct Library*)IntuitionBase);
      ConsoleOpened = FALSE;
   }
}


   /*
    *    Print a String to the Console
    */

void Puts(char *str)
{
   if (ConsoleOpened) {
      ConsoleRequest->io_Command = CMD_WRITE;
      ConsoleRequest->io_Length = -1;
      ConsoleRequest->io_Data = str;
      SendIO((struct IORequest *)ConsoleRequest);
      if (Aux)
         while (*str)
            RawPutChar(*str++);
      WaitIO((struct IORequest *)ConsoleRequest);
   }
}


   /*
    *    Put a Character to the Console
    */

void PutChar(char c)
{
   if (ConsoleOpened) {
      ConsoleRequest->io_Command = CMD_WRITE;
      ConsoleRequest->io_Length = 1;
      ConsoleRequest->io_Data = &c;
      SendIO((struct IORequest *)ConsoleRequest);
      if (Aux)
         RawPutChar(c);
      WaitIO((struct IORequest *)ConsoleRequest);
   }
}


   /*
    *    Get a Character from the Console
    */

long GetChar(void)
{
   long res = -1;
   char c;

   if (ConsoleOpened) {
      ConsoleRequest->io_Command = CMD_READ;
      ConsoleRequest->io_Length = 1;
      ConsoleRequest->io_Data = &c;
      SendIO((struct IORequest *)ConsoleRequest);
      do
         if (Aux && ((res = RawMayGetChar()) != -1))
            break;
         else if (CheckIO((struct IORequest *)ConsoleRequest)) {
            res = c;
            break;
         }
      while (MasterMode || Debug || !TimedOut());
      AbortIO((struct IORequest *)ConsoleRequest);
      WaitIO((struct IORequest *)ConsoleRequest);
   }
   return(res);
}


   /*
    *    Format and Print a String to the Console
    */

asm("
.text
PutCharInBuf:  movel    %a0,%sp@-
               lea      %pc@(FmtBuffer+255),%a0
               cmpl     %a0,%a3
               jne      NotFull
               clrb     %a3@
               jra      1f
NotFull:       moveb    %d0,%a3@+
1:             movel    %sp@+,%a0
               rts
");
extern void *PutCharInBuf(char c);

void Printf(char *fmt, ...)
{
   va_list args;

   va_start(args, fmt);
   RawDoFmt(fmt, args, (void (*)())PutCharInBuf, FmtBuffer);
   Puts(FmtBuffer);
   va_end(args);
}
   

   /*
    *    Read a Line from the Console
    */

static long ReadLine(u_long mode)
{
   long c;
   long length = 0;
   u_long eol = FALSE;

   while (!eol && ((c = GetChar()) != -1))
      switch (c) {
         case '\r':
         case '\n':
            InBuffer[length] = '\0';
            if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
               PutChar('\n');
            eol = TRUE;
            break;
         case '\b':
            if (length > 0) {
               length--;
               if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
                  Puts("\b \b");
            }
            break;
         default:
            if (length < sizeof(InBuffer)-1) {
               InBuffer[length++] = c;
               switch (mode) {
                  case ECHO_NORMAL:
                     PutChar(c);
                     break;
                  case ECHO_DOT:
                     PutChar('+');
                     break;
               }
            }
            break;
      }
   return(eol ? length : -1);
}


   /*
    *    Set the Time Out Value
    */

static void SetTimeOut(u_long seconds)
{
   if (!TimeOutSignalMask) {
      TimerRequest->io_Command = TR_ADDREQUEST;
      TimerRequest->tv_secs = seconds;
      TimerRequest->tv_micro = 0;
      SendIO((struct IORequest *)TimerRequest);
      TimeOutSignalMask = 1<<TimerPort->mp_SigBit;
   }
}


   /*
    *    Check for a Time Out
    */

static int TimedOut(void)
{
   int res = FALSE;

   if (TimeOutSignalMask && CheckIO((struct IORequest *)TimerRequest))
      res = TRUE;

   return(res);
}


   /*
    *    Clear Time Out
    */

static void ClearTimeOut(void)
{
   if (TimeOutSignalMask) {
      AbortIO((struct IORequest *)TimerRequest);
      WaitIO((struct IORequest *)TimerRequest);
      TimeOutSignalMask = 0;
   }
}


   /*
    *    Sleep for a Specified Period
    */

void Sleep(u_long micros)
{
   TimerRequest->io_Command = TR_ADDREQUEST;
   TimerRequest->io_Flags = IOF_QUICK;
   TimerRequest->tv_secs = micros/1000000;
   TimerRequest->tv_micro = micros%1000000;
   DoIO((struct IORequest *)TimerRequest);
}


   /*
    *    Get the Default Boot Operating System
    */

static struct MapEntry *GetDefaultBootOS(void)
{
   struct MapEntry *bootentry = MapEntries;
   int i;

   for (i = 0; (i < MapHeader.Default) && bootentry; i++)
      bootentry = bootentry->un.Next;
   if (!bootentry)
      Alert(AN_LILO|AO_LiloMap);

   return(bootentry);
}


   /*
    *    Ask the User for a Boot Operating System
    */

static struct MapEntry *GetBootOS(void)
{
   struct MapEntry *bootentry = GetDefaultBootOS();
   struct MapEntry *entry;
   long type;

   if (MapHeader.TimeOut)
      SetTimeOut(MapHeader.TimeOut);

   while (1) {
      Puts("\nBoot image: ");
      if ((ReadLine(ECHO_NORMAL)) < 1)
         break;
      if (!strcmp(InBuffer, "help") || !strcmp(InBuffer, "?")) {
         Puts("\nValid commands: su, help, ?\n\nList of boot images:\n\n");
         for (entry = MapEntries; entry; entry = entry->un.Next) {
            type = entry->Type;
            if (type < 0)
               type = -type;
            if (entry == bootentry)
               Puts("[1m");
            Printf("   %-31.31s (%s%s%s)\n", entry->Label,
                   type == BOS_AMIGA ? "AmigaOS" :
                   type == BOS_LINUX ? "Linux/m68k" :
                   type == BOS_NETBSD ? "NetBSD-Amiga" : "Unknown",
                   entry->Password[0] ? ", Restricted" : "",
                   entry->Type < 0 ? ", N/A" : "");
            if (entry == bootentry)
               Puts("[0m");
         }
      } else if (!strcmp(InBuffer, "su")) {
         if (MasterMode)
            continue;
         if (MapHeader.MasterPwd[0]) {
            Puts("\nPassword: ");
            if (((ReadLine(ECHO_DOT)) < 1) ||
                strcmp(InBuffer, MapHeader.MasterPwd)) {
               Puts("\nIncorrect password.\n");
               continue;
            }
         }
         MasterMode = TRUE;
         Puts("\nEntering Master Mode.\n");
      } else {
         for (entry = MapEntries; entry; entry = entry->un.Next)
            if (!strcmp(InBuffer, entry->Label))
               break;
         if (!entry) {
            Printf("\nNonexistent boot image `%s'.\n", InBuffer);
            continue;
         } else if (entry->Type < 0) {
            Printf("\nBoot image `%s' is not available.\n", InBuffer);
            continue;
         } else if (!MasterMode && entry->Password[0]) {
            Puts("\n\nPassword: ");
            if (((ReadLine(ECHO_DOT)) < 1) ||
                strcmp(InBuffer, entry->Password)) {
               Puts("\nIncorrect password.\n");
               continue;
            }
         }
         bootentry = entry;
         break;
      }
   }

   if (MasterMode) {
      Printf("\n\nOld command line: %s\n\nNew command line: ",
             bootentry->CommandLine);
      ReadLine(ECHO_NORMAL);
      strncpy(bootentry->CommandLine, InBuffer, CL_SIZE);
      bootentry->CommandLine[CL_SIZE-1] = '\0';
   }

   if (MapHeader.TimeOut)
      ClearTimeOut();

   return(bootentry);
}


   /*
    *    Boot The Standard Amiga Operating System
    */

static u_long BootAmigaOS(void)
{
   return(TRUE);
}


   /*
    *    Functions needed by LinuxBoot
    *
    *    These functions are not complete implementations. They do only what's
    *    really necessary.
    */

int Open(char *name)
{
   Lilo_Open(BootEntry->KernelVector);
   return(TRUE);
}


int Seek(int fd, int offset, int whence)
{
   Lilo_Seek(offset);
   return(offset);
}


int Read(int fd, char *buf, int count)
{
   Lilo_Read(buf, count);
   return(count);
}


void Close(int fd)
{
   Lilo_Close();
}


   /*
    *    Modify the bootinfo, if necessary
    */

int ModifyBootInfo(struct bootinfo *bi)
{
   int i;

   if (Debug)
      Puts("Modifying bootinfo structure\n");

   if (BootEntry->ChipSize) {
      bi->bi_amiga.chip_size = BootEntry->ChipSize;
      if (Debug)
         Printf("Chip RAM: %10ld bytes\n", BootEntry->ChipSize);
   }

   for (i = 0; i < BootEntry->NumFastRAM && i < NUM_MEMINFO; i++) {
      bi->memory[i].addr = BootEntry->FastAddress[i];
      bi->memory[i].size = BootEntry->FastSize[i];
      if (Debug)
         Printf("Fast RAM: %10ld bytes at 0x%08lx\n", BootEntry->FastSize[i],
                BootEntry->FastAddress[i]);
   }
   if (i != bi->num_memory && i > 0)
      bi->num_memory = i;

   if (BootEntry->Model != AMI_UNKNOWN) {
      bi->bi_amiga.model = BootEntry->Model;
      if (Debug)
         Printf("Model   : #%ld\n", BootEntry->Model);
   }

   return(TRUE);
}
